/************************************************************
 *                                                          *
 *            8-channel programmable party flasher          * 
 *                                                          *
 *  author  : (c) 2025, mateusko O.A.M.D.G                  *
 *            mateusko.oamdg@outlook.com                    *
 *  date    : 2024-01-15                                    *
 *  version : 0.0.8                                         *
 *  microcontroler: Arduino UNO R3                          *
  **********************************************************/

/*
 *  8-channel programmable party flasher is free software: 
 *  you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/gpl.html>
 */

#define IR_RECEIVE_PIN      10 
#define RAW_BUFFER_LENGTH  400

/* PRISPÔSOBENIE PROGRAMU PRE VLASTNÝ INFRAČERVENÝ OVLÁDAČ
 *
 * Číslo protokolu a kódy tlačidiel vášho infračerveného ovládača 
 * získate spustením tohoto skeču.
 * Po stlačení tlačidla na IRC sa v Sériovom Monitore vypíše 
 * číslo protokolu a kód tlačidla. Napríklad:
 *
 *      Protocol: 8   (NEC)
 *      Code: 0xFE010202
 * 
 * Číslo protokolu a kódy tlačidiel vložíte tu nižšie:
 */

#define READED_PROTOCOL -1 // -1 = všetky protokoly

#define IR_BTN_1 0xFE010202     // Scenár blikania č. 1
#define IR_BTN_2 0xFD020202     // Scenár blikania č. 2
#define IR_BTN_3 0xFC030202     // Scenár blikania č. 3
#define IR_BTN_4 0xFB040202     // Scenár blikania č. 4
#define IR_BTN_5 0xFA050202     // Scenár blikania č. 5
#define IR_BTN_6 0xF9060202     // Scenár blikania č. 6
#define IR_BTN_7 0xF8070202     // Scenár blikania č. 7
#define IR_BTN_8 0xF7080202     // Scenár blikania č. 8
#define IR_BTN_9 0xF6090202     // Scenár blikania č. 9
#define IR_BTN_0 0xFF000202     // Scenár blikania č. 0

#define IR_BTN_PLUS  0xF40B0202 // Zvýšenie jasu svietenia
#define IR_BTN_MINUS 0xF10E0202 // Zníženie jasu svietenia
#define IR_BTN_OFF   0xF50A0202 // Vypnutie

#if !defined(STR_HELPER)
#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)
#endif

#define IR_USE_AVR_TIMER1

#include "avr/pgmspace.h"
#include <IRremote.hpp>


/* Scenáre blikania sú v textových reťazcoch, ktoré sú zložené z príkazov:


    +       nasledujúce čísla kanálov LED sa zapnú
    -       nasledujúce čísla kanálov LED sa vypnú
    0..7    čísla kanálov LED
    $       všetky kanály LED sa zapnú (skrátený zápis +01234567)
    @       všetky kanály LED sa vypnú (skrátený zápis -01234567)

    A..H    pauza, dĺžka páuz sa dá nastaviť 
    S[A..H]<cas v ms> - (SET) nastavenie pauzy, <unsigned long> je celočíselná hodnota v ms
            napr. SA1000   - nastavenie pauzy A na 1000 milisekúnd  1 s
            napr. SB200    - nastavenie pauzy B na 200 milisekúnd = 0,2 s
    
    L (LOOP)    L<pocet opakovani>(<prikazy>) 
            napr. L5(+1A -1A)  - 5 krát sa zopakuje: zapne kanál 1, počká sa pauzu A
                                                     vypne sa kanál 1 a počká sa pauzu A

    R (REPEAT)  skok na značku M alebo ak nie je uvedená tak skok na začiatok scenára
    
    M (MARK)    značka, kam má program skočiť pri opakovaní
    
    medzi príkazy môžete vkladať medzery pre lepšiu čitateľnosť
    neznáme znaky sa ignorujú
*/ 

// Trvalé svietenie všetkých kanálov LED
const char scene1[] PROGMEM = "$";

// Pomalé prepínanie svietenia 0-4 a 5-7
const char scene2[] PROGMEM = "SA2000 M @ +01234 A -01234 +567 A R";

// "posun" viacerých doprava
const char scene3[] PROGMEM = "SA500 M @ +036 A @ +147 A @ +25 A R";

// Príklad zložitejšieho scenára
const char scene4[] PROGMEM = {
  "@SA150 SB200 SC400 SD1000 "
  "+0C-0+1C-1+2C-2+3C-3+4C-4C"
  "+012345"
  "L18(+56A-56A) "
  "L5(+01234B-01234B) "
  "L5(+56B-56B) "
  "L3($A@B) B"
  "+0C1C2C3C4C5C6C -0C1C2C3C4C5C6C @"
  "+06C-06+15C-15+24C-24+3C+24C+15C+06C-3B-24B-15B-06B"
  "L3($A@B) B"
  "+56 D @"
  "+01234 D @"
  "L5(+0 +56A-56A -0+1 56A-56A -1+2 +56A-56A -2+3 +56A-56A -3+4 +56A-56A -4) R"};

// 0-4 blikajú postupne doprava, 567 trvalo svietia
const char scene5[] PROGMEM = {"SA100 @ +567 M +0A -0+1A -1+2A -2+3A -3+4A -4+3A -3+2A -2+1A -1 R"};

// 3x zabliká 0-4, potom 3x zabliká 5-7
const char scene6[] PROGMEM = {"@ SA100 SB500 L3 (+01234 A -01234 A) B L3(+567 A -567 A) BB R"};

// 5x Posun doprava potom 5x posun doľava
const char scene7[] PROGMEM = {"SA50 M @ L5(+0 A -7+1 A -0+2 A -1+3 A -2+4 A -3+5 A -4+6 A -5+7A -6)"
                               "       @ L5(+7 A -0+6 A +5-7 A +4-6 A +3-5 A +2-4 A +1-3 A +0-2 A -1) R" };

// Príklad zložitejšieho scenára
const char scene8[] PROGMEM = {"SA50 @ M +0A1A2A34A-0A1A2A3A4A +5-6 +0A1A2A3A4A-0A1A2A3A4 -5+6A R"};

// Príklad zložitejšieho scenára
const char scene9[] PROGMEM = {
  "SA50 SB100 SC200 SD500 @  L10($D@B) M"
  "L10(+0A -0+1A -1+2A -2+3A -3+4A   +0A -0+1A -1+2A -2+3A    +0A -0+1A -1+2A    +0A -0+1A +0C  "
  "     -4A -3+4A -2+3A -1+2A -0+1A  -4A -3+4A -2+3A -1+2A    -4A -3+4A -2+3A    -4A -3+4A -4B)  "
  "-0A -1A -2A -3A -4A L18(+56B-56B)   "
  "L10($C@B) +56"
  "+56 L20(+04 A -04+13A -13+2A -2A +2A -2A +2A -2A +13A -13+04) $D"
  "@ L20(-6+5B -5+6B)@ D R"
  };

// Protiposun  
const char scene0[] PROGMEM = {"SA50 @ M +07 A -07+16 A -16+25 A -25+34 A -43+52 A -52+61 A -61 R"};

const char sceneOFF[] PROGMEM = {"@"}; // vypnutie všetkých kanálov

// Nastavenie pinov mikrokontroléra pre kanály 0 - 7
int ledPin[8] = {2, 3, 4, 5, 6, 7, 8, 9};             


// Premenné

unsigned long pause[8];    // Dĺžka páuz v ms (index 0 je pauza A, index 7 je pauza H
int  state = LOW;          // aktuálny status pre kanály LOW=vypnúť, HIGH=zapnúť
char* scene;               // pointer na aktuálny scenár blikania
bool isPause = false; 
bool isPlaying = false;
int startLoopIndex = 0;
int markIndex = 0;
int loopCounter = 0;
int index;                // práve čítaný znak v aktuálnom scenári  
unsigned long sceneTimer;
unsigned long delayTime;
float brightness  = 1.0;  // intenzita svietenia


// Funkcie

// Príjem 1 bytového príkazu z infračerveného ovládača
char readIR(); 

// Interpret scenára blikania
void play(const char* scene);   // spustenie "prehratia" scenára 
void sceneUpdate();             // treba volať neustále v cykle

// PRIVATE, don't use
bool endOfScene(); //returns true if end of scene occured
char readChar();     //reads character from scene
bool nextChar();     //move index to next char (index++)
unsigned long  readULInt();      //reads and returns unsigned long integer from scene
void initialize();                //initialize pause presets
void setLedStatus(byte ledIndex, int status); //switching LED[ledIndex] to state 'status'
const byte brigs[] = { 251, 248, 244, 234, 220,180, 160,128,64,32,1,0};
int brigs_index = 5;

// Prednastavené dĺžky páuz
void initialize() {
  pause[0] = 10;      // A
  pause[1] = 20;      // B
  pause[2] = 50;      // C
  pause[3] = 100;     // D
  pause[4] = 200;     // E
  pause[5] = 500;     // F
  pause[6] = 1000;    // G 
  pause[7] = 2000;    // H
}

// Vracia true, ak je koniec scenára; inak vracia false
bool endOfScene() {
  return readChar() == 0;
}

// Prečíta a vráti jeden znak zo scenára
char readChar() {
  return pgm_read_byte_near(scene + index);
}

// Posunie kurzor na ďalší znak v scenári a vráti true, ak nie je koniec scenára
// Zároveň preskakuje medzery
bool nextChar() {
  if (readChar() == 0) return false;
  index ++;
  while(readChar() == ' ') index++;
  return readChar() != 0;
}

// Nastavenie pauzy. 
// Písmeno pauzy a čas trvania sa číta z aktuálnej pozície v scenári.
void readAndPreset() {
  int index = 0; 
  char c = readChar();

  if (endOfScene()) return;

  index = c - 'A';

  if (index >= 0 && index <= 7) {
    nextChar();
    if (endOfScene()) return;
    unsigned long x = readULInt();
    pause[index] = x;
    Serial.print(F("> SET PAUSE "));
    Serial.print(c);
    Serial.print(F(" to "));
    Serial.print(x, DEC);
    Serial.println(F(" ms"));
  }
}

// Čítanie celočíselnej hodnoty z aktuálnej pozície scenára
unsigned long readULInt() {
  unsigned long x = 0;
  
  while (!endOfScene()) {
    char c = readChar();
    if (c >= '0' && c <= '9') {
        x *= 10;
        x += c - '0';
        nextChar();
    }
    else break;
  }

  return x;
}

// Fyzické zapnutie alebo vypnutie kanála ledIndex
void setLedStatus(byte ledIndex, int status) {
  int pin = ledPin[ledIndex];
  if (pin > 0) digitalWrite(pin, status);
}

// Štart "prehrávania0" scenára
void play(const char* sceneX) {
  initialize();
  state = LOW;
  index = 0;
  startLoopIndex = 0;
  loopCounter = 0;
  markIndex = 0;
  scene = (char*) sceneX;
  isPause = false;
  isPlaying = true;
  sceneTimer = millis();
  delayTime = 0;
}

// Aktualizácia stavu scenára - hlavná výkonná funkcia
void sceneUpdate() {
  bool next = true;
  char c;

  if (!isPlaying) return;

  // Pauza 
  if (isPause) {
    if (millis() - sceneTimer > delayTime) {
      isPause = false;
    }
    else {
      return; 
    }
  }

  // End of scene
  if (endOfScene()) {
    isPlaying = false;
    return;
  }

  // Read command  
  c = readChar();

  // Command "+"
  if (c == '+') {
    state = HIGH;
    Serial.println(F("> SET STATE = HIGH"));
  }

  // Command "-"
  else if (c == '-') {
    state = LOW;
    Serial.println(F("> SET STATE = LOW"));
  }

  // Command "0".."7"
  else if (c >= '0' && c <= '7') {
    int pin = ledPin[c - '0'];
    if (pin > 0) {
      digitalWrite(pin, state);
      
      Serial.print(F("> SET LED "));
      Serial.print(c - '0', DEC);
      Serial.print(F(" to "));
      if (state) 
        Serial.println(F("HIGH"));
      else 
        Serial.println(F("LOW"));
    }
  }

  // Commands "A".."H" (Pause)
  else if (c >= 'A' && c <= 'H') {
    delayTime = pause[c - 'A'];
    sceneTimer = millis();
    isPause = true;
    
    Serial.print(F("> PAUSE("));
    Serial.print(c);
    Serial.print(F(") "));
    Serial.print(pause[c - 'A'], DEC);
    Serial.println(F(" ms"));
  }

  // Command "S" (Set)
  else if (c == 'S') {
    nextChar();
    readAndPreset();  //set pause array item via scene
    next = false;
  }
    
  // Command "$" (All On)
  else if (c == '$') {
    for (int i = 0; i < 8; i++) setLedStatus(i, HIGH);
    Serial.println(F("> ALL LEDs = HIGH"));
  }

  // Command "@" (All Off)
  else if (c == '@') {
    for (int i = 0; i < 8; i++) setLedStatus(i, LOW);
    Serial.println(F("> ALL LEDs = LOW"));
  }

  // Command "R" (Repeat from begin or from mark M)
  else if (c == 'R') {
    index = markIndex;
    next = false;
  }

  // Command "L" (Loop) 
  else if (c == 'L') {
    if (!nextChar()) return;
    loopCounter = readULInt();
    if (loopCounter > 0) loopCounter--;
    Serial.print(F("SET Loop Counter = "));
    Serial.println(loopCounter + 1, DEC);
    // --- Start Loop
    if(endOfScene()) return;
    if(readChar() == '(') {
      if (!nextChar()) return;
      Serial.print(F("SET startLoopIndex = "));
      Serial.println(index, DEC); 
      startLoopIndex = index;
      next = false;
    }
  }
    // --- End Loop  
  else if (c == ')') {
    if (loopCounter > 0) {
      loopCounter--;
      index = startLoopIndex;
      next = false;
    }
  }

  else if (c == 'M') {
    markIndex = index;
  }
  
  if (next) nextChar();
}


void setup() {
  
  // Inicializácia Sériového rozhrania pre výpisy
  Serial.begin(115200);

  // Inicializácia IR prijímača a dekódera
  IrReceiver.begin(IR_RECEIVE_PIN);
  Serial.print(F("Ready to receive IR signals of protocols: "));
  printActiveIRProtocols(&Serial);
  
  // Inicializácia výstupných pinov mikrokontroléra pre kanály 0 .. 7
  for (int i = 0; i < 8; i++) pinMode(ledPin[i], OUTPUT);

  // Inicializácia pinu PWM - intenzita svietenia
  pinMode(11, OUTPUT);
  analogWrite(11, brigs[brigs_index]);

  // Blink on start
  for (int i = 0; i < 8; i++) digitalWrite(ledPin[i], HIGH);delay(100);for (int i = 0; i < 8; i++)  digitalWrite(ledPin[i], LOW);  
}

// Príjem signálu z IR ovládača a vykonanie príslušnej akcie
void doIRCommand() {
  switch (readIR()) {
    case '+':
      brigs_index++;
      if (brigs[brigs_index] == 0 ) brigs_index--;
      analogWrite(11, brigs[brigs_index]);
      Serial.print(F("> JAS = ")); Serial.println(byte(brightness));
      break;

    case '-':
      if (brigs_index > 0 ) brigs_index--;
      analogWrite(11, brigs[brigs_index]);
      Serial.print(F("> JAS = ")); Serial.println(byte(brightness));
      break;

    case 'F':
      play(sceneOFF);

      break;

    case '1':
      play(scene1);
      break;

    case '2':
      play(scene2);
      break;

    case '3':
      play(scene3);
      break;

    case '4':
      play(scene4);
      break;

    case '5':
      play(scene5);
      break;

    case '6':
      play(scene6);
      break;

    case '7':
      play(scene7);
      break;

     case '8':
      play(scene8);
      break;

     case '9':
      play(scene9);
      break;

     case '0':
      play(scene0);
      break;
  }
}


void loop() {
  
  // Handle Infrared Remote Controller commands
  doIRCommand(); 

  // Update blink LEDs
  sceneUpdate();
}


// Prevod 32bitového kódu IRC na 1-bytový príkaz
char irButton(unsigned long rawData) {
  char value = 0;
  switch (rawData) {
    case IR_BTN_1: value = '1'; break; 
    case IR_BTN_2: value = '2'; break;
    case IR_BTN_3: value =  '3'; break;
    case IR_BTN_4: value =  '4'; break;
    case IR_BTN_5: value =  '5'; break;
    case IR_BTN_6: value =  '6'; break;
    case IR_BTN_7: value =  '7'; break;
    case IR_BTN_8: value =  '8'; break;
    case IR_BTN_9: value =  '9'; break;
    case IR_BTN_0: value =  '0'; break;
    case IR_BTN_PLUS: value =  '+'; break;
    case IR_BTN_MINUS: value =  '-'; break;
    case IR_BTN_OFF: value =  'F'; break;
  };
  
  return value; 
}

// Fyzický príjem kódu z infračerveného ovládača ako 32bitové číslo
uint32_t receiveIR() {
  static uint32_t old_hash = 0;
  static uint32_t timer = 0;
  
  uint32_t hash;
  if (IrReceiver.decode()) {
        IrReceiver.resume();

        hash = IrReceiver.decodedIRData.decodedRawData;
        if (hash == old_hash) {
          if (millis() - timer < 200u) {
            timer = millis();
            return 0;
          } 
        }
        
        timer = millis();
        old_hash = hash;
        return hash;
        
    }
    return 0;   
}

// Príjem kódu z IRC a dekódovanie na 1 bytový príkaz
// 
char readIR() {
  uint32_t ircode = receiveIR();
  
  // Vypis prijateho kodu od IRC
  if (ircode > 0) {
    Serial.print(F("Protocol: "));
    Serial.print(IrReceiver.decodedIRData.protocol);
    Serial.print(F("  ("));
    Serial.print(getProtocolString(IrReceiver.decodedIRData.protocol));
    Serial.println(')');
    Serial.print(F("Code: 0x"));
    Serial.println(ircode, HEX);
    Serial.println();
  }

  if (READED_PROTOCOL < 0 || IrReceiver.decodedIRData.protocol == READED_PROTOCOL)
    return irButton(ircode); 
  else
    return 0;
}

// Inicializacia IR prijímača a dekódera
void initIR(){
  IrReceiver.begin(IR_RECEIVE_PIN);
  Serial.print(F("Ready to receive IR signals of protocols: "));
  printActiveIRProtocols(&Serial);
}